"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.keystone
"""

import pytest

import salt.modules.config as config
import salt.modules.keystone as keystone
from tests.support.mock import MagicMock, call, patch


class MockEC2:
    """
    Mock of EC2 class
    """

    def __init__(self):
        self.access = ""
        self.secret = ""
        self.tenant_id = ""
        self.user_id = ""
        self.connection_args = ""
        self.profile = ""

    @staticmethod
    def create(userid, tenantid):
        """
        Mock of create method
        """
        cr_ec2 = MockEC2()
        cr_ec2.tenant_id = tenantid
        cr_ec2.user_id = userid
        return cr_ec2

    def delete(self, userid, accesskey):
        """
        Mock of delete method
        """
        self.access = accesskey
        self.user_id = userid
        return True

    @staticmethod
    def get(user_id, access, profile, **connection_args):
        """
        Mock of get method
        """
        cr_ec2 = MockEC2()
        cr_ec2.profile = profile
        cr_ec2.access = access
        cr_ec2.user_id = user_id
        cr_ec2.connection_args = connection_args
        return cr_ec2

    @staticmethod
    def list(user_id):
        """
        Mock of list method
        """
        cr_ec2 = MockEC2()
        cr_ec2.user_id = user_id
        return [cr_ec2]


class MockEndpoints:
    """
    Mock of Endpoints class
    """

    def __init__(self):
        self.id = "007"
        self.region = "RegionOne"
        self.adminurl = "adminurl"
        self.internalurl = "internalurl"
        self.publicurl = "publicurl"
        self.service_id = "117"

    @staticmethod
    def list():
        """
        Mock of list method
        """
        return [MockEndpoints()]

    @staticmethod
    def create(region, service_id, publicurl, adminurl, internalurl):
        """
        Mock of create method
        """
        return (region, service_id, publicurl, adminurl, internalurl)

    @staticmethod
    def delete(id):
        """
        Mock of delete method
        """
        return id


class MockServices:
    """
    Mock of Services class
    """

    flag = None

    def __init__(self):
        self.id = "117"
        self.name = "iptables"
        self.description = "description"
        self.type = "type"

    @staticmethod
    def create(name, service_type, description):
        """
        Mock of create method
        """
        service = MockServices()
        service.id = "005"
        service.name = name
        service.description = description
        service.type = service_type
        return service

    def get(self, service_id):
        """
        Mock of get method
        """
        service = MockServices()
        if self.flag == 1:
            service.id = "asd"
            return [service]
        elif self.flag == 2:
            service.id = service_id
            return service
        return [service]

    def list(self):
        """
        Mock of list method
        """
        service = MockServices()
        if self.flag == 1:
            service.id = "asd"
            return [service]
        return [service]

    @staticmethod
    def delete(service_id):
        """
        Mock of delete method
        """
        return service_id


class MockRoles:
    """
    Mock of Roles class
    """

    flag = None

    def __init__(self):
        self.id = "113"
        self.name = "nova"
        self.user_id = "446"
        self.tenant_id = "a1a1"

    @staticmethod
    def create(name):
        """
        Mock of create method
        """
        return name

    def get(self, role_id):
        """
        Mock of get method
        """
        role = MockRoles()
        if self.flag == 1:
            role.id = None
            return role
        role.id = role_id
        return role

    @staticmethod
    def list():
        """
        Mock of list method
        """
        return [MockRoles()]

    @staticmethod
    def delete(role):
        """
        Mock of delete method
        """
        return role

    @staticmethod
    def add_user_role(user_id, role_id, tenant_id):
        """
        Mock of add_user_role method
        """
        return (user_id, role_id, tenant_id)

    @staticmethod
    def remove_user_role(user_id, role_id, tenant_id):
        """
        Mock of remove_user_role method
        """
        return (user_id, role_id, tenant_id)

    @staticmethod
    def roles_for_user(user, tenant):
        """
        Mock of roles_for_user method
        """
        role = MockRoles()
        role.user_id = user
        role.tenant_id = tenant
        return [role]


class MockTenants:
    """
    Mock of Tenants class
    """

    flag = None

    def __init__(self):
        self.id = "446"
        self.name = "nova"
        self.description = "description"
        self.enabled = "True"

    @staticmethod
    def create(name, description, enabled):
        """
        Mock of create method
        """
        tenant = MockTenants()
        tenant.name = name
        tenant.description = description
        tenant.enabled = enabled
        return tenant

    def get(self, tenant_id):
        """
        Mock of get method
        """
        tenant = MockTenants()
        if self.flag == 1:
            tenant.id = None
            return tenant
        tenant.id = tenant_id
        return tenant

    @staticmethod
    def list():
        """
        Mock of list method
        """
        return [MockTenants()]

    @staticmethod
    def delete(tenant_id):
        """
        Mock of delete method
        """
        return tenant_id


class MockServiceCatalog:
    """
    Mock of ServiceCatalog class
    """

    def __init__(self):
        self.id = "446"
        self.expires = "No"
        self.user_id = "admin"
        self.tenant_id = "ae04"

    def get_token(self):
        """
        Mock of get_token method
        """
        return {
            "id": self.id,
            "expires": self.expires,
            "user_id": self.user_id,
            "tenant_id": self.tenant_id,
        }


class MockUsers:
    """
    Mock of Users class
    """

    flag = None

    def __init__(self):
        self.id = "446"
        self.name = "nova"
        self.email = "salt@saltstack.com"
        self.enabled = "True"
        self.tenant_id = "a1a1"
        self.password = "salt"

    def create(self, name, password, email, tenant_id, enabled):
        """
        Mock of create method
        """
        user = MockUsers()
        user.name = name
        user.password = password
        user.email = email
        user.enabled = enabled
        self.tenant_id = tenant_id
        return user

    def get(self, user_id):
        """
        Mock of get method
        """
        user = MockUsers()
        if self.flag == 1:
            user.id = None
            return user
        user.id = user_id
        return user

    @staticmethod
    def list():
        """
        Mock of list method
        """
        return [MockUsers()]

    @staticmethod
    def delete(user_id):
        """
        Mock of delete method
        """
        return user_id

    @staticmethod
    def update(user, name, email, enabled):
        """
        Mock of update method
        """
        return (user, name, email, enabled)

    @staticmethod
    def update_password(user, password):
        """
        Mock of update_password method
        """
        return (user, password)


class Unauthorized(Exception):
    """
    The base exception class for all exceptions.
    """

    def __init__(self, message="Test"):
        super().__init__(message)
        self.msg = message


class AuthorizationFailure(Exception):
    """
    Additional exception class to Unauthorized.
    """

    def __init__(self, message="Test"):
        super().__init__(message)
        self.msg = message


class MockExceptions:
    """
    Mock of exceptions class
    """

    def __init__(self):
        self.Unauthorized = Unauthorized
        self.AuthorizationFailure = AuthorizationFailure


class MockKeystoneClient:
    """
    Mock of keystoneclient module
    """

    def __init__(self):
        self.exceptions = MockExceptions()


class MockClient:
    """
    Mock of Client class
    """

    flag = None

    def __init__(self, profile=None, **conn_args):
        self.ec2 = MockEC2()
        self.endpoints = MockEndpoints()
        self.services = MockServices()
        self.roles = MockRoles()
        self.tenants = MockTenants()
        self.service_catalog = MockServiceCatalog()
        self.users = MockUsers()

    def Client(self, **kwargs):
        """
        Mock of Client method
        """
        if self.flag == 1:
            raise Unauthorized
        return True


@pytest.fixture
def configure_loader_modules():
    return {
        keystone: {
            "auth": MockClient,
            "client": MockClient(),
            "keystoneclient": MockKeystoneClient(),
            "__salt__": {"config.get": config.get},
            "__opts__": {},
        },
        config: {"__opts__": {}},
    }


def test_ec2_credentials_create():
    """
    Test if it create EC2-compatible credentials for user per tenant
    """
    assert keystone.ec2_credentials_create() == {"Error": "Could not resolve User ID"}

    assert keystone.ec2_credentials_create(user_id="salt") == {
        "Error": "Could not resolve Tenant ID"
    }

    assert keystone.ec2_credentials_create(user_id="salt", tenant_id="72278") == {
        "access": "",
        "tenant_id": "72278",
        "secret": "",
        "user_id": "salt",
    }


def test_ec2_credentials_delete():
    """
    Test if it delete EC2-compatible credentials
    """
    assert keystone.ec2_credentials_delete() == {"Error": "Could not resolve User ID"}

    assert (
        keystone.ec2_credentials_delete(user_id="salt", access_key="72278")
        == 'ec2 key "72278" deleted under user id "salt"'
    )


def test_ec2_credentials_get():
    """
    Test if it return ec2_credentials for a user
    (keystone ec2-credentials-get)
    """
    assert keystone.ec2_credentials_get() == {"Error": "Unable to resolve user id"}

    assert keystone.ec2_credentials_get(user_id="salt") == {
        "Error": "Access key is required"
    }

    assert keystone.ec2_credentials_get(
        user_id="salt", access="72278", profile="openstack1"
    ) == {
        "salt": {
            "access": "72278",
            "secret": "",
            "tenant": "",
            "user_id": "salt",
        }
    }


def test_ec2_credentials_list():
    """
    Test if it return a list of ec2_credentials
    for a specific user (keystone ec2-credentials-list)
    """
    assert keystone.ec2_credentials_list() == {"Error": "Unable to resolve user id"}

    assert keystone.ec2_credentials_list(user_id="salt", profile="openstack1") == {
        "salt": {"access": "", "secret": "", "tenant_id": "", "user_id": "salt"}
    }


def test_endpoint_get():
    """
    Test if it return a specific endpoint (keystone endpoint-get)
    """
    assert keystone.endpoint_get("nova", "RegionOne", profile="openstack") == {
        "Error": "Could not find the specified service"
    }

    ret = {"Error": "Could not find endpoint for the specified service"}
    MockServices.flag = 1
    assert keystone.endpoint_get("iptables", "RegionOne", profile="openstack") == ret

    MockServices.flag = 0
    assert keystone.endpoint_get("iptables", "RegionOne", profile="openstack") == {
        "adminurl": "adminurl",
        "id": "007",
        "internalurl": "internalurl",
        "publicurl": "publicurl",
        "region": "RegionOne",
        "service_id": "117",
    }


def test_endpoint_list():
    """
    Test if it return a list of available endpoints
    (keystone endpoints-list)
    """
    assert keystone.endpoint_list(profile="openstack1") == {
        "007": {
            "adminurl": "adminurl",
            "id": "007",
            "internalurl": "internalurl",
            "publicurl": "publicurl",
            "region": "RegionOne",
            "service_id": "117",
        }
    }


def test_endpoint_create():
    """
    Test if it create an endpoint for an Openstack service
    """
    assert keystone.endpoint_create("nova") == {
        "Error": "Could not find the specified service"
    }

    MockServices.flag = 2
    assert keystone.endpoint_create(
        "iptables",
        "http://public/url",
        "http://internal/url",
        "http://adminurl/url",
        "RegionOne",
    ) == {
        "adminurl": "adminurl",
        "id": "007",
        "internalurl": "internalurl",
        "publicurl": "publicurl",
        "region": "RegionOne",
        "service_id": "117",
    }


def test_endpoint_delete():
    """
    Test if it delete an endpoint for an Openstack service
    """
    ret = {"Error": "Could not find any endpoints for the service"}
    assert keystone.endpoint_delete("nova", "RegionOne") == ret

    with patch.object(
        keystone, "endpoint_get", MagicMock(side_effect=[{"id": "117"}, None])
    ):
        assert keystone.endpoint_delete("iptables", "RegionOne")


def test_role_create():
    """
    Test if it create named role
    """
    assert keystone.role_create("nova") == {"Error": 'Role "nova" already exists'}

    assert keystone.role_create("iptables") == {"Error": "Unable to resolve role id"}


def test_role_delete():
    """
    Test if it delete a role (keystone role-delete)
    """
    assert keystone.role_delete() == {"Error": "Unable to resolve role id"}

    assert keystone.role_delete("iptables") == "Role ID iptables deleted"


def test_role_get():
    """
    Test if it return a specific roles (keystone role-get)
    """
    assert keystone.role_get() == {"Error": "Unable to resolve role id"}

    assert keystone.role_get(name="nova") == {"nova": {"id": "113", "name": "nova"}}


def test_role_list():
    """
    Test if it return a list of available roles (keystone role-list)
    """
    assert keystone.role_list() == {
        "nova": {
            "id": "113",
            "name": "nova",
            "tenant_id": "a1a1",
            "user_id": "446",
        }
    }


def test_service_create():
    """
    Test if it add service to Keystone service catalog
    """
    MockServices.flag = 2
    assert keystone.service_create("nova", "compute", "OpenStack Service") == {
        "iptables": {
            "description": "description",
            "id": "005",
            "name": "iptables",
            "type": "type",
        }
    }


def test_service_delete():
    """
    Test if it delete a service from Keystone service catalog
    """
    assert (
        keystone.service_delete("iptables") == 'Keystone service ID "iptables" deleted'
    )


def test_service_get():
    """
    Test if it return a list of available services (keystone services-list)
    """
    MockServices.flag = 0
    assert keystone.service_get() == {"Error": "Unable to resolve service id"}

    MockServices.flag = 2
    assert keystone.service_get(service_id="c965") == {
        "iptables": {
            "description": "description",
            "id": "c965",
            "name": "iptables",
            "type": "type",
        }
    }


def test_service_list():
    """
    Test if it return a list of available services (keystone services-list)
    """
    MockServices.flag = 0
    assert keystone.service_list(profile="openstack1") == {
        "iptables": {
            "description": "description",
            "id": "117",
            "name": "iptables",
            "type": "type",
        }
    }


def test_tenant_create():
    """
    Test if it create a keystone tenant
    """
    assert keystone.tenant_create("nova") == {
        "nova": {
            "description": "description",
            "id": "446",
            "name": "nova",
            "enabled": "True",
        }
    }


def test_tenant_delete():
    """
    Test if it delete a tenant (keystone tenant-delete)
    """
    assert keystone.tenant_delete() == {"Error": "Unable to resolve tenant id"}

    assert keystone.tenant_delete("nova") == "Tenant ID nova deleted"


def test_tenant_get():
    """
    Test if it return a specific tenants (keystone tenant-get)
    """
    assert keystone.tenant_get() == {"Error": "Unable to resolve tenant id"}

    assert keystone.tenant_get(tenant_id="446") == {
        "nova": {
            "description": "description",
            "id": "446",
            "name": "nova",
            "enabled": "True",
        }
    }


def test_tenant_list():
    """
    Test if it return a list of available tenants (keystone tenants-list)
    """
    assert keystone.tenant_list() == {
        "nova": {
            "description": "description",
            "id": "446",
            "name": "nova",
            "enabled": "True",
        }
    }


def test_tenant_update():
    """
    Test if it update a tenant's information (keystone tenant-update)
    """
    assert keystone.tenant_update() == {"Error": "Unable to resolve tenant id"}


def test_token_get():
    """
    Test if it return the configured tokens (keystone token-get)
    """
    assert keystone.token_get() == {
        "expires": "No",
        "id": "446",
        "tenant_id": "ae04",
        "user_id": "admin",
    }


def test_user_list():
    """
    Test if it return a list of available users (keystone user-list)
    """
    assert keystone.user_list() == {
        "nova": {
            "name": "nova",
            "tenant_id": "a1a1",
            "enabled": "True",
            "id": "446",
            "password": "salt",
            "email": "salt@saltstack.com",
        }
    }


def test_user_get():
    """
    Test if it return a specific users (keystone user-get)
    """
    assert keystone.user_get() == {"Error": "Unable to resolve user id"}

    assert keystone.user_get(user_id="446") == {
        "nova": {
            "name": "nova",
            "tenant_id": "a1a1",
            "enabled": "True",
            "id": "446",
            "password": "salt",
            "email": "salt@saltstack.com",
        }
    }


def test_user_create():
    """
    Test if it create a user (keystone user-create)
    """
    assert keystone.user_create(
        name="nova",
        password="salt",
        email="salt@saltstack.com",
        tenant_id="a1a1",
    ) == {
        "nova": {
            "name": "nova",
            "tenant_id": "a1a1",
            "enabled": "True",
            "id": "446",
            "password": "salt",
            "email": "salt@saltstack.com",
        }
    }


def test_user_delete():
    """
    Test if it delete a user (keystone user-delete)
    """
    assert keystone.user_delete() == {"Error": "Unable to resolve user id"}

    assert keystone.user_delete("nova") == "User ID nova deleted"


def test_user_update():
    """
    Test if it update a user's information (keystone user-update)
    """
    assert keystone.user_update() == {"Error": "Unable to resolve user id"}

    assert keystone.user_update("nova") == "Info updated for user ID nova"


def test_user_verify_password():
    """
    Test if it verify a user's password
    """
    mock = MagicMock(return_value="http://127.0.0.1:35357/v2.0")
    with patch.dict(keystone.__salt__, {"config.option": mock}):
        assert keystone.user_verify_password() == {
            "Error": "Unable to resolve user name"
        }

        assert keystone.user_verify_password(user_id="446", name="nova")

        MockClient.flag = 1
        assert not keystone.user_verify_password(user_id="446", name="nova")


def test_user_password_update():
    """
    Test if it update a user's password (keystone user-password-update)
    """
    assert keystone.user_password_update() == {"Error": "Unable to resolve user id"}

    assert keystone.user_password_update("nova") == "Password updated for user ID nova"


def test_user_role_add():
    """
    Test if it add role for user in tenant (keystone user-role-add)
    """
    assert (
        keystone.user_role_add(user="nova", tenant="nova", role="nova")
        == '"nova" role added for user "nova" for "nova" tenant/project'
    )

    MockRoles.flag = 1
    assert keystone.user_role_add(user="nova", tenant="nova", role="nova") == {
        "Error": "Unable to resolve role id"
    }

    MockTenants.flag = 1
    assert keystone.user_role_add(user="nova", tenant="nova") == {
        "Error": "Unable to resolve tenant/project id"
    }

    MockUsers.flag = 1
    assert keystone.user_role_add(user="nova") == {"Error": "Unable to resolve user id"}


def test_user_role_remove():
    """
    Test if it add role for user in tenant (keystone user-role-add)
    """
    MockUsers.flag = 1
    assert keystone.user_role_remove(user="nova") == {
        "Error": "Unable to resolve user id"
    }

    MockUsers.flag = 0
    MockTenants.flag = 1
    assert keystone.user_role_remove(user="nova", tenant="nova") == {
        "Error": "Unable to resolve tenant/project id"
    }

    MockTenants.flag = 0
    MockRoles.flag = 1
    assert keystone.user_role_remove(user="nova", tenant="nova", role="nova") == {
        "Error": "Unable to resolve role id"
    }

    ret = '"nova" role removed for user "nova" under "nova" tenant'
    MockRoles.flag = 0
    assert keystone.user_role_remove(user="nova", tenant="nova", role="nova") == ret


def test_user_role_list():
    """
    Test if it return a list of available user_roles
    (keystone user-roles-list)
    """
    assert keystone.user_role_list(user="nova") == {
        "Error": "Unable to resolve user or tenant/project id"
    }

    assert keystone.user_role_list(user_name="nova", tenant_name="nova") == {
        "nova": {
            "id": "113",
            "name": "nova",
            "tenant_id": "446",
            "user_id": "446",
        }
    }


def test_api_version_verify_ssl():
    """
    test api_version when using verify_ssl
    """
    test_verify = [True, False, None]
    conn_args = {
        "keystone.user": "admin",
        "connection_password": "password",
        "connection_tenant": "admin",
        "connection_tenant_id": "id",
        "connection_auth_url": "https://127.0.0.1/v2.0/",
        "connection_verify_ssl": True,
    }

    http_ret = {"dict": {"version": {"id": "id_test"}}}
    for verify in test_verify:
        mock_http = MagicMock(return_value=http_ret)
        patch_http = patch("salt.utils.http.query", mock_http)
        conn_args["connection_verify_ssl"] = verify
        if verify is None:
            conn_args.pop("connection_verify_ssl")
            verify = True

        with patch_http:
            ret = keystone.api_version(**conn_args)

        assert mock_http.call_args_list == [
            call(
                "https://127.0.0.1/v2.0/",
                decode=True,
                decode_type="json",
                verify_ssl=verify,
            )
        ]



"""
Integration tests for the match runner
"""

import logging

import pytest

from tests.conftest import FIPS_TESTRUN

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
]


@pytest.fixture(scope="class")
def match_master_config():
    return {
        "open_mode": True,
        "peer_run": {
            "match-minion-bob": [
                "match.compound_matches",
            ],
        },
        "nodegroups": {
            "alice_eve": "I@name:ali*",
            "alice": "L@match-minion-alice",
        },
        "minion_data_cache": True,
    }


@pytest.fixture(scope="class", autouse=True)
def pillar_tree(match_salt_master, match_salt_minion_alice, match_salt_minion_eve):
    top_file = f"""
    base:
      '{match_salt_minion_alice.id}':
        - alice
      '{match_salt_minion_eve.id}':
        - eve
    """
    alice_pillar_file = """
    name: alice
    """
    eve_pillar_file = """
    name: alice_whoops_sorry_eve_hrhr
    """
    top_tempfile = match_salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    alice_tempfile = match_salt_master.pillar_tree.base.temp_file(
        "alice.sls", alice_pillar_file
    )
    eve_tempfile = match_salt_master.pillar_tree.base.temp_file(
        "eve.sls", eve_pillar_file
    )

    with top_tempfile, alice_tempfile, eve_tempfile:
        ret = match_salt_minion_alice.salt_call_cli().run(
            "saltutil.refresh_pillar", wait=True
        )
        assert ret.returncode == 0
        assert ret.data is True
        ret = match_salt_minion_eve.salt_call_cli().run(
            "saltutil.refresh_pillar", wait=True
        )
        assert ret.returncode == 0
        assert ret.data is True
        yield


@pytest.fixture(scope="class")
def match_salt_master(salt_factories, match_master_config):
    factory = salt_factories.salt_master_daemon(
        "match-master",
        defaults=match_master_config,
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "publish_signing_algorithm": (
                "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
            ),
        },
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="class")
def match_salt_minion_alice(match_salt_master):
    assert match_salt_master.is_running()
    factory = match_salt_master.salt_minion_daemon(
        "match-minion-alice",
        defaults={"open_mode": True, "grains": {"role": "alice"}},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def match_salt_minion_eve(match_salt_master):
    assert match_salt_master.is_running()
    factory = match_salt_master.salt_minion_daemon(
        "match-minion-eve",
        defaults={"open_mode": True, "grains": {"role": "eve"}},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def match_salt_minion_bob(match_salt_master):
    assert match_salt_master.is_running()
    factory = match_salt_master.salt_minion_daemon(
        "match-minion-bob",
        defaults={"open_mode": True},
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def match_salt_call_cli(match_salt_minion_bob):
    return match_salt_minion_bob.salt_call_cli()


@pytest.fixture(scope="class")
def match_salt_run_cli(match_salt_master):
    return match_salt_master.salt_run_cli()


class TestMatchCompoundRunner:
    @pytest.fixture
    def alice_uncached(self, match_salt_minion_alice, match_salt_run_cli):
        ret = match_salt_run_cli.run("cache.clear_all", "match-minion-alice")
        assert ret.returncode == 0
        yield
        match_salt_minion_alice.salt_call_cli().run("pillar.items")

    @pytest.fixture
    def eve_cached(self, match_salt_minion_eve):
        ret = match_salt_minion_eve.salt_call_cli().run("pillar.items")
        assert ret.returncode == 0
        yield

    @pytest.fixture
    def alice_down(self, match_salt_minion_alice):
        with match_salt_minion_alice.stopped():
            yield

    @pytest.mark.parametrize(
        "expr,expected",
        [
            ("G@role:alice", True),
            ("G@role:ali*", True),
            (r"E@match\-minion\-(alice|bob)", True),
            ("P@role:^(alice|bob)$", True),
            ("L@match-minion-alice,match-minion-bob", True),
            ("I@name:alice", True),
            ("I@name:ali*", False),
            ("J@name:alice", True),
            ("J@name:^(alice|bob)$", False),
            ("N@alice", True),
            ("N@alice_eve", False),
            ("G@role:ali* and I@name:alice", True),
            ("G@role:ali* and I@name:ali*", False),
            ("G@role:ali* or I@name:ali*", True),
            ("G@role:ali* and not I@name:alice", False),
        ],
    )
    def test_match_compound_matches(self, match_salt_run_cli, expr, expected):
        if expected:
            expected = "match-minion-alice"
        ret = match_salt_run_cli.run(
            "match.compound_matches", expr, "match-minion-alice"
        )
        assert ret.returncode == 0
        assert ret.data
        assert "res" in ret.data
        assert ret.data["res"] == expected

    @pytest.mark.usefixtures("eve_cached")
    def test_match_compound_matches_only_allows_exact_pillar_matching(
        self, match_salt_run_cli
    ):
        """
        This check is mostly redundant, but better check explicitly with the scenario
        to prevent because it is security-critical.
        """
        ret = match_salt_run_cli.run(
            "match.compound_matches", "I@name:alic*", "match-minion-eve"
        )
        assert ret.returncode == 0
        assert ret.data
        assert "res" in ret.data
        assert ret.data["res"] is False

    @pytest.mark.parametrize(
        "expr,expected",
        [
            ("G@role:alice", False),
            ("G@role:ali*", False),
            ("I@name:alice", False),
            ("I@name:ali*", False),
            ("G@role:ali* and I@name:alice", False),
            ("L@match-minion-alice,match-minion-bob", True),
            ("L@match-minion-alice,match-minion-bob and G@role:alice", False),
            ("L@match-minion-alice,match-minion-bob and I@name:alice", False),
        ],
    )
    @pytest.mark.usefixtures("alice_uncached")
    def test_match_compound_matches_with_uncached_minion_data(
        self, match_salt_run_cli, expr, expected
    ):
        if expected:
            expected = "match-minion-alice"
        ret = match_salt_run_cli.run(
            "match.compound_matches", expr, "match-minion-alice"
        )
        assert ret.returncode == 0
        assert ret.data
        assert "res" in ret.data
        assert ret.data["res"] == expected

    @pytest.mark.parametrize(
        "expr,expected",
        [
            ("G@role:alice", True),
            ("G@role:ali*", True),
            ("I@name:alice", True),
            ("I@name:ali*", False),
            ("G@role:ali* and I@name:alice", True),
            ("L@match-minion-alice,match-minion-bob", True),
        ],
    )
    @pytest.mark.usefixtures("alice_down")
    def test_match_compound_matches_when_minion_is_down(
        self, match_salt_run_cli, expr, expected
    ):
        if expected:
            expected = "match-minion-alice"
        ret = match_salt_run_cli.run(
            "match.compound_matches", expr, "match-minion-alice"
        )
        assert ret.returncode == 0
        assert ret.data
        assert "res" in ret.data
        assert ret.data["res"] == expected

    @pytest.mark.parametrize(
        "minion_id",
        [
            "hi\\there",
            "my/minion",
            "../../../../../../../../../etc/shadow",
        ],
    )
    def test_match_compound_matches_with_invalid_minion_id(
        self, minion_id, match_salt_run_cli
    ):
        ret = match_salt_run_cli.run("match.compound_matches", "*", minion_id)
        assert ret.returncode == 0
        assert ret.data
        assert "res" in ret.data
        assert ret.data["res"] is False

    @pytest.mark.parametrize(
        "expr,expected",
        [
            ("G@role:alice", True),
            ("G@role:ali*", True),
            ("I@name:alice", True),
            ("I@name:ali*", False),
            ("G@role:ali* and I@name:alice", True),
            ("L@match-minion-alice,match-minion-bob", True),
        ],
    )
    def test_match_compound_matches_as_peer_run(
        self, match_salt_call_cli, expr, expected
    ):
        if expected:
            expected = "match-minion-alice"
        ret = match_salt_call_cli.run(
            "publish.runner", "match.compound_matches", [expr, "match-minion-alice"]
        )
        assert ret.returncode == 0
        assert ret.data
        assert "res" in ret.data
        assert ret.data["res"] == expected


class TestMatchCompoundRunnerWithoutMinionDataCache:
    @pytest.fixture(scope="class")
    def match_master_config(self):
        return {
            "open_mode": True,
            "minion_data_cache": False,
        }

    @pytest.mark.parametrize(
        "expr",
        [
            "G@role:alice",
            "G@role:ali*",
            "I@name:alice",
            "I@name:ali*",
            "G@role:ali* and I@name:alice",
        ],
    )
    def test_match_compound_matches(self, match_salt_run_cli, expr):
        ret = match_salt_run_cli.run(
            "match.compound_matches", expr, "match-minion-alice"
        )
        assert ret.returncode == 0
        assert ret.data
        assert "res" in ret.data
        assert ret.data["res"] is False



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.guestfs
"""

import pytest

import salt.modules.guestfs as guestfs
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {guestfs: {}}


# 'mount' function tests: 1
def test_mount():
    """
    Test if it mounts an image
    """
    # Test case with non-existing mount folder
    run_mock = MagicMock(return_value="")
    with patch("os.path.join", MagicMock(return_value="/tmp/guest/fedora.qcow")), patch(
        "os.path.isdir", MagicMock(return_value=False)
    ), patch("os.makedirs", MagicMock()) as makedirs_mock, patch(
        "os.listdir", MagicMock(return_value=False)
    ), patch.dict(
        guestfs.__salt__, {"cmd.run": run_mock}
    ):
        assert guestfs.mount("/srv/images/fedora.qcow")
        run_mock.assert_called_once_with(
            "guestmount -i -a /srv/images/fedora.qcow --rw /tmp/guest/fedora.qcow",
            python_shell=False,
        )
        makedirs_mock.assert_called_once()

    # Test case with existing but empty mount folder
    run_mock.reset_mock()
    with patch("os.path.join", MagicMock(return_value="/tmp/guest/fedora.qcow")), patch(
        "os.path.isdir", MagicMock(return_value=True)
    ), patch("os.makedirs", MagicMock()) as makedirs_mock, patch(
        "os.listdir", MagicMock(return_value=False)
    ), patch.dict(
        guestfs.__salt__, {"cmd.run": run_mock}
    ):
        assert guestfs.mount("/srv/images/fedora.qcow")
        run_mock.assert_called_once_with(
            "guestmount -i -a /srv/images/fedora.qcow --rw /tmp/guest/fedora.qcow",
            python_shell=False,
        )
        makedirs_mock.assert_not_called()

    # Test case with existing but not empty mount folder
    run_mock.reset_mock()
    with patch(
        "os.path.join",
        MagicMock(side_effect=["/tmp/guest/fedora.qcow", "/tmp/guest/fedora.qcowabc"]),
    ), patch("os.path.isdir", MagicMock(side_effect=[True, False])), patch(
        "os.makedirs", MagicMock()
    ) as makedirs_mock, patch(
        "os.listdir", MagicMock(side_effect=[True, False])
    ), patch.dict(
        guestfs.__salt__, {"cmd.run": run_mock}
    ):
        assert guestfs.mount("/srv/images/fedora.qcow")
        run_mock.assert_called_once_with(
            "guestmount -i -a /srv/images/fedora.qcow --rw"
            " /tmp/guest/fedora.qcowabc",
            python_shell=False,
        )
        makedirs_mock.assert_called_once()


@pytest.mark.slow_test
def test_umount():
    """
    Test the guestfs.unmount function
    """
    run_mock = MagicMock(side_effect=["", "lsof output line", ""])
    with patch.dict(guestfs.__salt__, {"cmd.run": run_mock}):
        guestfs.umount("/tmp/mnt/opensuse.qcow", disk="/path/to/opensuse.qcow")
        expected = [
            call("guestunmount -q /tmp/mnt/opensuse.qcow"),
            call("lsof /path/to/opensuse.qcow"),
            call("lsof /path/to/opensuse.qcow"),
        ]
        assert expected == run_mock.call_args_list

    # Test without the disk image path
    run_mock = MagicMock(side_effect=[""])
    with patch.dict(guestfs.__salt__, {"cmd.run": run_mock}):
        guestfs.umount("/tmp/mnt/opensuse.qcow")
        expected = [call("guestunmount -q /tmp/mnt/opensuse.qcow")]
        assert expected == run_mock.call_args_list



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import pytest

import salt.modules.genesis as genesis
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {genesis: {}}


def test_bootstrap():
    """
    Test for Create an image for a specific platform.
    """
    exception_string = "Exception({})".format(repr("foo"))
    mock = MagicMock(return_value=False)
    with patch.dict(genesis.__salt__, {"file.directory_exists": mock}):
        mock = MagicMock(side_effect=Exception("foo"))
        with patch.dict(genesis.__salt__, {"file.mkdir": mock}):
            assert genesis.bootstrap("platform", "root") == {"Error": exception_string}

    with patch.object(genesis, "_bootstrap_yum", return_value="A"):
        with patch.dict(
            genesis.__salt__,
            {
                "mount.umount": MagicMock(),
                "file.rmdir": MagicMock(),
                "file.directory_exists": MagicMock(),
            },
        ):
            with patch.dict(
                genesis.__salt__, {"disk.blkid": MagicMock(return_value={})}
            ):
                assert genesis.bootstrap("rpm", "root", "dir") is None

    common_parms = {
        "platform": "deb",
        "root": "root",
        "img_format": "dir",
        "arch": "amd64",
        "flavor": "stable",
        "static_qemu": "qemu",
    }

    param_sets = [
        {
            "params": {},
            "cmd": [
                "debootstrap",
                "--foreign",
                "--arch",
                "amd64",
                "stable",
                "root",
                "http://ftp.debian.org/debian/",
            ],
        },
        {
            "params": {"pkgs": "vim"},
            "cmd": [
                "debootstrap",
                "--foreign",
                "--arch",
                "amd64",
                "--include",
                "vim",
                "stable",
                "root",
                "http://ftp.debian.org/debian/",
            ],
        },
        {
            "params": {"pkgs": "vim,emacs"},
            "cmd": [
                "debootstrap",
                "--foreign",
                "--arch",
                "amd64",
                "--include",
                "vim,emacs",
                "stable",
                "root",
                "http://ftp.debian.org/debian/",
            ],
        },
        {
            "params": {"pkgs": ["vim", "emacs"]},
            "cmd": [
                "debootstrap",
                "--foreign",
                "--arch",
                "amd64",
                "--include",
                "vim,emacs",
                "stable",
                "root",
                "http://ftp.debian.org/debian/",
            ],
        },
        {
            "params": {"pkgs": ["vim", "emacs"], "exclude_pkgs": ["vim", "foo"]},
            "cmd": [
                "debootstrap",
                "--foreign",
                "--arch",
                "amd64",
                "--include",
                "vim,emacs",
                "--exclude",
                "vim,foo",
                "stable",
                "root",
                "http://ftp.debian.org/debian/",
            ],
        },
    ]

    for param_set in param_sets:

        with patch.dict(
            genesis.__salt__,
            {
                "mount.umount": MagicMock(),
                "file.rmdir": MagicMock(),
                "file.directory_exists": MagicMock(),
                "cmd.run": MagicMock(),
                "disk.blkid": MagicMock(return_value={}),
            },
        ):
            with patch("salt.modules.genesis.salt.utils.path.which", return_value=True):
                with patch(
                    "salt.modules.genesis.salt.utils.validate.path.is_executable",
                    return_value=True,
                ):
                    param_set["params"].update(common_parms)
                    assert genesis.bootstrap(**param_set["params"]) is None
                    genesis.__salt__["cmd.run"].assert_any_call(
                        param_set["cmd"], python_shell=False
                    )

    with patch.object(genesis, "_bootstrap_pacman", return_value="A") as pacman_patch:
        with patch.dict(
            genesis.__salt__,
            {
                "mount.umount": MagicMock(),
                "file.rmdir": MagicMock(),
                "file.directory_exists": MagicMock(),
                "disk.blkid": MagicMock(return_value={}),
            },
        ):
            genesis.bootstrap("pacman", "root", "dir")
            pacman_patch.assert_called_with(
                "root", img_format="dir", exclude_pkgs=[], pkgs=[]
            )


def test_avail_platforms():
    """
    Test for Return which platforms are available
    """
    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        assert not genesis.avail_platforms()["deb"]


def test_pack():
    """
    Test for Pack up a directory structure, into a specific format
    """
    with patch.object(genesis, "_tar", return_value="tar"):
        assert genesis.pack("name", "root") is None


def test_unpack():
    """
    Test for Unpack an image into a directory structure
    """
    with patch.object(genesis, "_untar", return_value="untar"):
        assert genesis.unpack("name", "root") is None



import pytest
from saltfactories.utils import random_string

from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def user(modules):
    return modules.user


@pytest.fixture
def username_str(user):
    _username = random_string("test-account-", uppercase=False)
    try:
        yield _username
    finally:
        try:
            user.delete(_username, purge=True, force=True)
        except Exception:  # pylint: disable=broad-except
            # The point here is just system cleanup. It can fail if no account was created
            pass


@pytest.fixture
def username_int(user):
    _username = random_string("", uppercase=False, lowercase=False, digits=True)
    try:
        yield _username
    finally:
        try:
            user.delete(_username, purge=True, force=True)
        except Exception:  # pylint: disable=broad-except
            # The point here is just system cleanup. It can fail if no account was created
            pass


@pytest.fixture
def account_str(user, username_str):
    with pytest.helpers.create_account(username=username_str) as account:
        user.addgroup(account.username, "Users")
        yield account


@pytest.fixture
def account_int(user, username_int):
    with pytest.helpers.create_account(username=username_int) as account:
        user.addgroup(account.username, "Users")
        yield account


def test_add_str(user, username_str):
    ret = user.add(name=username_str)
    assert ret is True
    assert username_str in user.list_users()


def test_add_int(user, username_int):
    ret = user.add(name=username_int)
    assert ret is True
    assert username_int in user.list_users()


def test_addgroup_str(user, account_str):
    ret = user.addgroup(account_str.username, "Backup Operators")
    assert ret is True
    ret = user.info(account_str.username)
    assert "Backup Operators" in ret["groups"]


def test_addgroup_int(user, account_int):
    ret = user.addgroup(account_int.username, "Backup Operators")
    assert ret is True
    ret = user.info(account_int.username)
    assert "Backup Operators" in ret["groups"]


def test_chfullname_str(user, account_str):
    ret = user.chfullname(account_str.username, "New Full Name")
    assert ret is True
    ret = user.info(account_str.username)
    assert ret["fullname"] == "New Full Name"


def test_chfullname_int(user, account_int):
    ret = user.chfullname(account_int.username, "New Full Name")
    assert ret is True
    ret = user.info(account_int.username)
    assert ret["fullname"] == "New Full Name"


def test_chgroups_single_str(user, account_str):
    groups = ["Backup Operators"]
    ret = user.chgroups(account_str.username, groups=groups)
    assert ret is True
    ret = user.info(account_str.username)
    groups.append("Users")
    assert sorted(ret["groups"]) == sorted(groups)


def test_chgroups_single_int(user, account_int):
    groups = ["Backup Operators"]
    ret = user.chgroups(account_int.username, groups=groups)
    assert ret is True
    ret = user.info(account_int.username)
    groups.append("Users")
    assert sorted(ret["groups"]) == sorted(groups)


def test_chgroups_list_str(user, account_str):
    groups = ["Backup Operators", "Guests"]
    ret = user.chgroups(account_str.username, groups=groups)
    assert ret is True
    ret = user.info(account_str.username)
    groups.append("Users")
    assert sorted(ret["groups"]) == sorted(groups)


def test_chgroups_list_int(user, account_int):
    groups = ["Backup Operators", "Guests"]
    ret = user.chgroups(account_int.username, groups=groups)
    assert ret is True
    ret = user.info(account_int.username)
    groups.append("Users")
    assert sorted(ret["groups"]) == sorted(groups)


def test_chgroups_list_append_false_str(user, account_str):
    groups = ["Backup Operators", "Guests"]
    ret = user.chgroups(account_str.username, groups=groups, append=False)
    assert ret is True
    ret = user.info(account_str.username)
    assert sorted(ret["groups"]) == sorted(groups)


def test_chgroups_list_append_false_int(user, account_int):
    groups = ["Backup Operators", "Guests"]
    ret = user.chgroups(account_int.username, groups=groups, append=False)
    assert ret is True
    ret = user.info(account_int.username)
    assert sorted(ret["groups"]) == sorted(groups)


def test_chhome_str(user, account_str):
    home = r"C:\spongebob\squarepants"
    ret = user.chhome(name=account_str.username, home=home)
    assert ret is True
    ret = user.info(name=account_str.username)
    assert ret["home"] == home


def test_chhome_int(user, account_int):
    home = r"C:\spongebob\squarepants"
    ret = user.chhome(name=account_int.username, home=home)
    assert ret is True
    ret = user.info(name=account_int.username)
    assert ret["home"] == home


def test_chprofile_str(user, account_str):
    profile = r"C:\spongebob\squarepants"
    ret = user.chprofile(name=account_str.username, profile=profile)
    assert ret is True
    ret = user.info(name=account_str.username)
    assert ret["profile"] == profile


def test_chprofile_int(user, account_int):
    profile = r"C:\spongebob\squarepants"
    ret = user.chprofile(name=account_int.username, profile=profile)
    assert ret is True
    ret = user.info(name=account_int.username)
    assert ret["profile"] == profile


def test_delete_str(user, account_str):
    ret = user.delete(name=account_str.username)
    assert ret is True
    assert user.info(name=account_str.username) == {}


def test_delete_int(user, account_int):
    ret = user.delete(name=account_int.username)
    assert ret is True
    assert user.info(name=account_int.username) == {}


def test_get_user_sid_str(user, account_str):
    ret = user.get_user_sid(account_str.username)
    assert ret.startswith("S-1-5")


def test_get_user_sid_int(user, account_int):
    ret = user.get_user_sid(account_int.username)
    assert ret.startswith("S-1-5")


def test_info_str(user, account_str):
    ret = user.info(account_str.username)
    assert ret["name"] == account_str.username
    assert ret["uid"].startswith("S-1-5")


def test_info_int(user, account_int):
    ret = user.info(account_int.username)
    assert ret["name"] == account_int.username
    assert ret["uid"].startswith("S-1-5")


def test_list_groups_str(user, account_str):
    ret = user.list_groups(account_str.username)
    assert ret == ["Users"]


def test_list_groups_int(user, account_int):
    ret = user.list_groups(account_int.username)
    assert ret == ["Users"]


def test_list_users(user, account_str):
    ret = user.list_users()
    assert account_str.username in ret


def test_removegroup_str(user, account_str):
    ret = user.removegroup(account_str.username, "Users")
    assert ret is True
    ret = user.info(account_str.username)
    assert ret["groups"] == []


def test_removegroup_int(user, account_int):
    ret = user.removegroup(account_int.username, "Users")
    assert ret is True
    ret = user.info(account_int.username)
    assert ret["groups"] == []


def test_rename_str(user, account_str):
    new_name = random_string("test-account-", uppercase=False)
    ret = user.rename(name=account_str.username, new_name=new_name)
    assert ret is True
    assert new_name in user.list_users()
    # Let's set it back so that it gets cleaned up...
    ret = user.rename(name=new_name, new_name=account_str.username)
    assert ret is True


def test_rename_str_missing(user, account_str):
    missing = random_string("test-account-", uppercase=False)
    with pytest.raises(CommandExecutionError):
        user.rename(name=missing, new_name="spongebob")


def test_rename_str_existing(user, account_str):
    new_existing = random_string("test-account-", uppercase=False)
    ret = user.add(name=new_existing)
    assert ret is True
    with pytest.raises(CommandExecutionError):
        user.rename(name=account_str.username, new_name=new_existing)
    # We need to clean this up because it wasn't created in a fixture
    ret = user.delete(name=new_existing, purge=True, force=True)
    assert ret is True
    assert new_existing not in user.list_users()


def test_rename_int(user, account_int):
    new_name = random_string("", uppercase=False, lowercase=False, digits=True)
    ret = user.rename(name=account_int.username, new_name=new_name)
    assert ret is True
    assert new_name in user.list_users()
    # Let's set it back so that it gets cleaned up...
    ret = user.rename(name=new_name, new_name=account_int.username)
    assert ret is True


def test_rename_int_missing(user, account_int):
    missing = random_string("", uppercase=False, lowercase=False, digits=True)
    with pytest.raises(CommandExecutionError):
        user.rename(name=missing, new_name="spongebob")


def test_rename_int_existing(user, account_int):
    new_existing = random_string("", uppercase=False, lowercase=False, digits=True)
    ret = user.add(name=new_existing)
    assert ret is True
    with pytest.raises(CommandExecutionError):
        user.rename(name=account_int.username, new_name=new_existing)
    # We need to clean this up because it wasn't created in a fixture
    ret = user.delete(name=new_existing, purge=True, force=True)
    assert ret is True
    assert new_existing not in user.list_users()


def test_setpassword_str(user, account_str):
    ret = user.setpassword(account_str.username, password="Sup3rS3cret")
    # We have no way of verifying the password was changed on Windows, so the
    # best we can do is check that the command completed successfully
    assert ret is True


def test_setpassword_int(user, account_int):
    ret = user.setpassword(account_int.username, password="Sup3rS3cret")
    # We have no way of verifying the password was changed on Windows, so the
    # best we can do is check that the command completed successfully
    assert ret is True


@pytest.mark.parametrize(
    "value_name, new_value, info_field, expected",
    [
        ("description", "New description", "", None),
        ("homedrive", "H:", "", None),
        ("logonscript", "\\\\server\\script.cmd", "", None),
        ("expiration_date", "3/19/2024", "", "2024-03-19 00:00:00"),
        ("expiration_date", "Never", "", None),
        ("account_disabled", True, "", None),
        ("account_disabled", False, "", None),
        ("unlock_account", True, "account_locked", False),
        ("password_never_expires", True, "", None),
        ("password_never_expires", False, "", None),
        ("expired", True, "", None),
        ("expired", False, "", None),
        ("disallow_change_password", True, "", None),
        ("disallow_change_password", False, "", None),
    ],
)
def test_update_str(user, value_name, new_value, info_field, expected, account_str):
    setting = {value_name: new_value}
    # You can't expire an account if the password never expires
    if value_name == "expired":
        setting.update({"password_never_expires": not new_value})
    ret = user.update(account_str.username, **setting)
    assert ret is True
    ret = user.info(account_str.username)
    info_field = info_field if info_field else value_name
    expected = expected if expected is not None else new_value
    assert ret[info_field] == expected



"""
Manage ruby gems.
"""

import logging
import re

import salt.utils.itertools
import salt.utils.platform
from salt.exceptions import CommandExecutionError

__func_alias__ = {"list_": "list"}

log = logging.getLogger(__name__)  # pylint: disable=C0103


def _gem(command, ruby=None, runas=None, gem_bin=None):
    """
    Run the actual gem command. If rvm or rbenv is installed, run the command
    using the corresponding module. rbenv is not available on windows, so don't
    try.

    :param command: string
    Command to run
    :param ruby: string : None
    If RVM or rbenv are installed, the ruby version and gemset to use.
    Ignored if ``gem_bin`` is specified.
    :param runas: string : None
    The user to run gem as.
    :param gem_bin: string : None
    Full path to the ``gem`` binary

    :return:
    Returns the full standard out including success codes or False if it fails
    """
    cmdline = [gem_bin or "gem"] + command

    # If a custom gem is given, use that and don't check for rvm/rbenv. User
    # knows best!
    if gem_bin is None:
        if __salt__["rvm.is_installed"](runas=runas):
            return __salt__["rvm.do"](ruby, cmdline, runas=runas)

        if not salt.utils.platform.is_windows() and __salt__["rbenv.is_installed"](
            runas=runas
        ):
            if ruby is None:
                return __salt__["rbenv.do"](cmdline, runas=runas)
            else:
                return __salt__["rbenv.do_with_ruby"](ruby, cmdline, runas=runas)

    ret = __salt__["cmd.run_all"](cmdline, runas=runas, python_shell=False)

    if ret["retcode"] == 0:
        return ret["stdout"]
    else:
        raise CommandExecutionError(ret["stderr"])


def install(
    gems,  # pylint: disable=C0103
    ruby=None,
    gem_bin=None,
    runas=None,
    version=None,
    rdoc=False,
    ri=False,
    pre_releases=False,
    proxy=None,
    source=None,
):  # pylint: disable=C0103
    """
    Installs one or several gems.

    :param gems: string
        The gems to install
    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.
    :param version: string : None
        Specify the version to install for the gem.
        Doesn't play nice with multiple gems at once
    :param rdoc: boolean : False
        Generate RDoc documentation for the gem(s).
        For rubygems > 3 this is interpreted as the --no-document arg and the
        ri option will then be ignored
    :param ri: boolean : False
        Generate RI documentation for the gem(s).
        For rubygems > 3 this is interpreted as the --no-document arg and the
        rdoc option will then be ignored
    :param pre_releases: boolean : False
        Include pre-releases in the available versions
    :param proxy: string : None
        Use the specified HTTP proxy server for all outgoing traffic.
        Format: http://hostname[:port]

    source : None
        Use the specified HTTP gem source server to download gem.
        Format: http://hostname[:port]

    CLI Example:

    .. code-block:: bash

        salt '*' gem.install vagrant

        salt '*' gem.install redphone gem_bin=/opt/sensu/embedded/bin/gem
    """
    try:
        gems = gems.split()
    except AttributeError:
        pass

    options = []
    if version:
        options.extend(["--version", version])
    if _has_rubygems_3(ruby=ruby, runas=runas, gem_bin=gem_bin):
        if not rdoc or not ri:
            options.append("--no-document")
        if pre_releases:
            options.append("--prerelease")
    else:
        if not rdoc:
            options.append("--no-rdoc")
        if not ri:
            options.append("--no-ri")
        if pre_releases:
            options.append("--pre")
    if proxy:
        options.extend(["-p", proxy])
    if source:
        options.extend(["--source", source])

    return _gem(["install"] + gems + options, ruby, gem_bin=gem_bin, runas=runas)


def uninstall(gems, ruby=None, runas=None, gem_bin=None):
    """
    Uninstall one or several gems.

    :param gems: string
        The gems to uninstall.
    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.uninstall vagrant
    """
    try:
        gems = gems.split()
    except AttributeError:
        pass

    return _gem(["uninstall"] + gems + ["-a", "-x"], ruby, gem_bin=gem_bin, runas=runas)


def update(gems, ruby=None, runas=None, gem_bin=None):
    """
    Update one or several gems.

    :param gems: string
        The gems to update.
    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.update vagrant
    """
    try:
        gems = gems.split()
    except AttributeError:
        pass

    return _gem(["update"] + gems, ruby, gem_bin=gem_bin, runas=runas)


def update_system(version="", ruby=None, runas=None, gem_bin=None):
    """
    Update rubygems.

    :param version: string : (newest)
        The version of rubygems to install.
    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.update_system
    """
    return _gem(["update", "--system", version], ruby, gem_bin=gem_bin, runas=runas)


def version(ruby=None, runas=None, gem_bin=None):
    """
    Print out the version of gem

    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.version
    """
    cmd = ["--version"]
    stdout = _gem(cmd, ruby, gem_bin=gem_bin, runas=runas)
    ret = {}
    for line in salt.utils.itertools.split(stdout, "\n"):
        match = re.match(r"[.0-9]+", line)
        if match:
            ret = line
            break
    return ret


def _has_rubygems_3(ruby=None, runas=None, gem_bin=None):
    match = re.match(r"^3\..*", version(ruby=ruby, runas=runas, gem_bin=gem_bin))
    if match:
        return True
    return False


def list_(prefix="", ruby=None, runas=None, gem_bin=None):
    """
    List locally installed gems.

    :param prefix: string :
        Only list gems when the name matches this prefix.
    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.list
    """
    cmd = ["list"]
    if prefix:
        cmd.append(prefix)
    stdout = _gem(cmd, ruby, gem_bin=gem_bin, runas=runas)
    ret = {}
    for line in salt.utils.itertools.split(stdout, "\n"):
        match = re.match(r"^([^ ]+) \((.+)\)", line)
        if match:
            gem = match.group(1)
            versions = match.group(2).split(", ")
            ret[gem] = versions
    return ret


def list_upgrades(ruby=None, runas=None, gem_bin=None):
    """
    .. versionadded:: 2015.8.0

    Check if an upgrade is available for installed gems

    gem_bin : None
        Full path to ``gem`` binary to use.
    ruby : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    runas : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.list_upgrades
    """
    result = _gem(["outdated"], ruby, gem_bin=gem_bin, runas=runas)
    ret = {}
    for line in salt.utils.itertools.split(result, "\n"):
        match = re.search(r"(\S+) \(\S+ < (\S+)\)", line)
        if match:
            name, version = match.groups()
        else:
            log.error("Can't parse line '%s'", line)
            continue
        ret[name] = version
    return ret


def sources_add(source_uri, ruby=None, runas=None, gem_bin=None):
    """
    Add a gem source.

    :param source_uri: string
        The source URI to add.
    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.sources_add http://rubygems.org/
    """
    return _gem(["sources", "--add", source_uri], ruby, gem_bin=gem_bin, runas=runas)


def sources_remove(source_uri, ruby=None, runas=None, gem_bin=None):
    """
    Remove a gem source.

    :param source_uri: string
        The source URI to remove.
    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.sources_remove http://rubygems.org/
    """
    return _gem(["sources", "--remove", source_uri], ruby, gem_bin=gem_bin, runas=runas)


def sources_list(ruby=None, runas=None, gem_bin=None):
    """
    List the configured gem sources.

    :param gem_bin: string : None
        Full path to ``gem`` binary to use.
    :param ruby: string : None
        If RVM or rbenv are installed, the ruby version and gemset to use.
        Ignored if ``gem_bin`` is specified.
    :param runas: string : None
        The user to run gem as.

    CLI Example:

    .. code-block:: bash

        salt '*' gem.sources_list
    """
    ret = _gem(["sources"], ruby, gem_bin=gem_bin, runas=runas)
    return [] if ret is False else ret.splitlines()[2:]



import logging
import os
import shutil
import subprocess
import tempfile
import urllib.error
import urllib.request

import pytest

import salt.modules.cmdmod as cmd
import salt.modules.virtualenv_mod
import salt.modules.zcbuildout as buildout
import salt.utils.files
import salt.utils.path
import salt.utils.platform
from tests.support.helpers import patched_environ
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
    pytest.mark.skip_on_windows(
        reason=(
            "Special steps are required for proper SSL validation because "
            "`easy_install` is too old(and deprecated)."
        )
    ),
]

KNOWN_VIRTUALENV_BINARY_NAMES = (
    "virtualenv",
    "virtualenv2",
    "virtualenv-2.6",
    "virtualenv-2.7",
)

BOOT_INIT = {
    1: ["var/ver/1/bootstrap/bootstrap.py"],
    2: ["var/ver/2/bootstrap/bootstrap.py", "b/bootstrap.py"],
}

log = logging.getLogger(__name__)


def download_to(url, dest):
    req = urllib.request.Request(url)
    req.add_header(
        "User-Agent",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36",
    )
    with salt.utils.files.fopen(dest, "wb") as fic:
        fic.write(urllib.request.urlopen(req, timeout=10).read())


class Base(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {
            buildout: {
                "__salt__": {
                    "cmd.run_all": cmd.run_all,
                    "cmd.run": cmd.run,
                    "cmd.retcode": cmd.retcode,
                }
            }
        }

    @classmethod
    def setUpClass(cls):
        if not os.path.isdir(RUNTIME_VARS.TMP):
            os.makedirs(RUNTIME_VARS.TMP)

        cls.root = os.path.join(RUNTIME_VARS.BASE_FILES, "buildout")
        cls.rdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        cls.tdir = os.path.join(cls.rdir, "test")
        for idx, url in buildout._URL_VERSIONS.items():
            log.debug("Downloading bootstrap from %s", url)
            dest = os.path.join(cls.rdir, f"{idx}_bootstrap.py")
            try:
                download_to(url, dest)
            except urllib.error.URLError as exc:
                log.debug("Failed to download %s: %s", url, exc)
        # creating a new setuptools install
        cls.ppy_st = os.path.join(cls.rdir, "psetuptools")
        if salt.utils.platform.is_windows():
            cls.bin_st = os.path.join(cls.ppy_st, "Scripts")
            cls.py_st = os.path.join(cls.bin_st, "python")
        else:
            cls.bin_st = os.path.join(cls.ppy_st, "bin")
            cls.py_st = os.path.join(cls.bin_st, "python")
        # `--no-site-packages` has been deprecated
        # https://virtualenv.pypa.io/en/stable/reference/#cmdoption-no-site-packages
        subprocess.check_call(
            [salt.utils.path.which_bin(KNOWN_VIRTUALENV_BINARY_NAMES), cls.ppy_st]
        )
        # Setuptools >=53.0.0 no longer has easy_install
        # Between 50.0.0 and 53.0.0 it has problems with salt, use an older version
        subprocess.check_call(
            [os.path.join(cls.bin_st, "pip"), "install", "-U", "setuptools<50.0.0"]
        )
        # distribute has been merged back in to setuptools as of v0.7. So, no
        # need to upgrade distribute, but this seems to be the only way to get
        # the binary in the right place
        # https://packaging.python.org/key_projects/#setuptools
        # Additionally, this part may fail if the certificate store is outdated
        # on Windows, as it would be in a fresh installation for example. The
        # following commands will fix that. This should be part of the golden
        # images. (https://github.com/saltstack/salt-jenkins/pull/1479)
        # certutil -generateSSTFromWU roots.sst
        # powershell "(Get-ChildItem -Path .\roots.sst) | Import-Certificate -CertStoreLocation Cert:\LocalMachine\Root"
        subprocess.check_call(
            [os.path.join(cls.bin_st, "easy_install"), "-U", "distribute"]
        )

    def setUp(self):
        if salt.utils.platform.is_darwin():
            self.patched_environ = patched_environ(__cleanup__=["__PYVENV_LAUNCHER__"])
            self.patched_environ.__enter__()  # pylint: disable=unnecessary-dunder-call
            self.addCleanup(self.patched_environ.__exit__)

        super().setUp()
        self._remove_dir()
        shutil.copytree(self.root, self.tdir)

        for idx in BOOT_INIT:
            path = os.path.join(self.rdir, f"{idx}_bootstrap.py")
            for fname in BOOT_INIT[idx]:
                shutil.copy2(path, os.path.join(self.tdir, fname))

    def tearDown(self):
        super().tearDown()
        self._remove_dir()

    def _remove_dir(self):
        if os.path.isdir(self.tdir):
            shutil.rmtree(self.tdir)


@pytest.mark.skip_if_binaries_missing(*KNOWN_VIRTUALENV_BINARY_NAMES, check_all=False)
@pytest.mark.requires_network
class BuildoutTestCase(Base):
    @pytest.mark.slow_test
    def test_onlyif_unless(self):
        b_dir = os.path.join(self.tdir, "b")
        ret = buildout.buildout(b_dir, onlyif=RUNTIME_VARS.SHELL_FALSE_PATH)
        self.assertTrue(ret["comment"] == "onlyif condition is false")
        self.assertTrue(ret["status"] is True)
        ret = buildout.buildout(b_dir, unless=RUNTIME_VARS.SHELL_TRUE_PATH)
        self.assertTrue(ret["comment"] == "unless condition is true")
        self.assertTrue(ret["status"] is True)

    @pytest.mark.slow_test
    def test_salt_callback(self):
        @buildout._salt_callback
        def callback1(a, b=1):
            for i in buildout.LOG.levels:
                getattr(buildout.LOG, i)(f"{i[0]}bar")
            return "foo"

        def callback2(a, b=1):
            raise Exception("foo")

        # pylint: disable=invalid-sequence-index
        ret1 = callback1(1, b=3)
        # These lines are throwing pylint errors - disabling for now since we are skipping
        # these tests
        # self.assertEqual(ret1['status'], True)
        # self.assertEqual(ret1['logs_by_level']['warn'], ['wbar'])
        # self.assertEqual(ret1['comment'], '')
        # These lines are throwing pylint errors - disabling for now since we are skipping
        # these tests
        # self.assertTrue(
        #     u''
        #     u'OUTPUT:\n'
        #     u'foo\n'
        #     u''
        #    in ret1['outlog']
        # )

        # These lines are throwing pylint errors - disabling for now since we are skipping
        # these tests
        # self.assertTrue(u'Log summary:\n' in ret1['outlog'])
        # These lines are throwing pylint errors - disabling for now since we are skipping
        # these tests
        # self.assertTrue(
        #     u'INFO: ibar\n'
        #     u'WARN: wbar\n'
        #     u'DEBUG: dbar\n'
        #     u'ERROR: ebar\n'
        #    in ret1['outlog']
        # )
        # These lines are throwing pylint errors - disabling for now since we are skipping
        # these tests
        # self.assertTrue('by level' in ret1['outlog_by_level'])
        # self.assertEqual(ret1['out'], 'foo')
        ret2 = buildout._salt_callback(callback2)(2, b=6)
        self.assertEqual(ret2["status"], False)
        self.assertTrue(ret2["logs_by_level"]["error"][0].startswith("Traceback"))
        self.assertTrue("Unexpected response from buildout" in ret2["comment"])
        self.assertEqual(ret2["out"], None)
        for l in buildout.LOG.levels:
            self.assertTrue(0 == len(buildout.LOG.by_level[l]))
        # pylint: enable=invalid-sequence-index

    @pytest.mark.slow_test
    def test_get_bootstrap_url(self):
        for path in [
            os.path.join(self.tdir, "var/ver/1/dumppicked"),
            os.path.join(self.tdir, "var/ver/1/versions"),
        ]:
            self.assertEqual(
                buildout._URL_VERSIONS[1],
                buildout._get_bootstrap_url(path),
                f"b1 url for {path}",
            )
        for path in [
            os.path.join(self.tdir, "/non/existing"),
            os.path.join(self.tdir, "var/ver/2/versions"),
            os.path.join(self.tdir, "var/ver/2/default"),
        ]:
            self.assertEqual(
                buildout._URL_VERSIONS[2],
                buildout._get_bootstrap_url(path),
                f"b2 url for {path}",
            )

    @pytest.mark.slow_test
    def test_get_buildout_ver(self):
        for path in [
            os.path.join(self.tdir, "var/ver/1/dumppicked"),
            os.path.join(self.tdir, "var/ver/1/versions"),
        ]:
            self.assertEqual(1, buildout._get_buildout_ver(path), f"1 for {path}")
        for path in [
            os.path.join(self.tdir, "/non/existing"),
            os.path.join(self.tdir, "var/ver/2/versions"),
            os.path.join(self.tdir, "var/ver/2/default"),
        ]:
            self.assertEqual(2, buildout._get_buildout_ver(path), f"2 for {path}")

    @pytest.mark.slow_test
    def test_get_bootstrap_content(self):
        self.assertEqual(
            "",
            buildout._get_bootstrap_content(os.path.join(self.tdir, "non", "existing")),
        )
        self.assertEqual(
            "",
            buildout._get_bootstrap_content(os.path.join(self.tdir, "var", "tb", "1")),
        )

        if (
            salt.utils.platform.is_windows()
            and os.environ.get("GITHUB_ACTIONS_PIPELINE", "0") == "0"
        ):
            line_break = "\r\n"
        else:
            line_break = "\n"
        self.assertEqual(
            f"# pylint: skip-file{line_break}foo{line_break}",
            buildout._get_bootstrap_content(os.path.join(self.tdir, "var", "tb", "2")),
        )

    @pytest.mark.slow_test
    def test_logger_clean(self):
        buildout.LOG.clear()
        # nothing in there
        self.assertTrue(
            True
            not in [len(buildout.LOG.by_level[a]) > 0 for a in buildout.LOG.by_level]
        )
        buildout.LOG.info("foo")
        self.assertTrue(
            True in [len(buildout.LOG.by_level[a]) > 0 for a in buildout.LOG.by_level]
        )
        buildout.LOG.clear()
        self.assertTrue(
            True
            not in [len(buildout.LOG.by_level[a]) > 0 for a in buildout.LOG.by_level]
        )

    @pytest.mark.slow_test
    def test_logger_loggers(self):
        buildout.LOG.clear()
        # nothing in there
        for i in buildout.LOG.levels:
            getattr(buildout.LOG, i)("foo")
            getattr(buildout.LOG, i)("bar")
            getattr(buildout.LOG, i)("moo")
            self.assertTrue(len(buildout.LOG.by_level[i]) == 3)
            self.assertEqual(buildout.LOG.by_level[i][0], "foo")
            self.assertEqual(buildout.LOG.by_level[i][-1], "moo")

    @pytest.mark.slow_test
    def test__find_cfgs(self):
        result = sorted(
            a.replace(self.root, "") for a in buildout._find_cfgs(self.root)
        )
        assertlist = sorted(
            [
                os.path.join(os.sep, "buildout.cfg"),
                os.path.join(os.sep, "c", "buildout.cfg"),
                os.path.join(os.sep, "etc", "buildout.cfg"),
                os.path.join(os.sep, "e", "buildout.cfg"),
                os.path.join(os.sep, "b", "buildout.cfg"),
                os.path.join(os.sep, "b", "bdistribute", "buildout.cfg"),
                os.path.join(os.sep, "b", "b2", "buildout.cfg"),
                os.path.join(os.sep, "foo", "buildout.cfg"),
            ]
        )
        self.assertEqual(result, assertlist)

    def skip_test_upgrade_bootstrap(self):
        b_dir = os.path.join(self.tdir, "b")
        bpy = os.path.join(b_dir, "bootstrap.py")
        buildout.upgrade_bootstrap(b_dir)
        time1 = os.stat(bpy).st_mtime
        with salt.utils.files.fopen(bpy) as fic:
            data = fic.read()
        self.assertTrue("setdefaulttimeout(2)" in data)
        flag = os.path.join(b_dir, ".buildout", "2.updated_bootstrap")
        self.assertTrue(os.path.exists(flag))
        buildout.upgrade_bootstrap(b_dir, buildout_ver=1)
        time2 = os.stat(bpy).st_mtime
        with salt.utils.files.fopen(bpy) as fic:
            data = fic.read()
        self.assertTrue("setdefaulttimeout(2)" in data)
        flag = os.path.join(b_dir, ".buildout", "1.updated_bootstrap")
        self.assertTrue(os.path.exists(flag))
        buildout.upgrade_bootstrap(b_dir, buildout_ver=1)
        time3 = os.stat(bpy).st_mtime
        self.assertNotEqual(time2, time1)
        self.assertEqual(time2, time3)


@pytest.mark.skip_if_binaries_missing(*KNOWN_VIRTUALENV_BINARY_NAMES, check_all=False)
@pytest.mark.requires_network
class BuildoutOnlineTestCase(Base):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.ppy_dis = os.path.join(cls.rdir, "pdistribute")
        cls.ppy_blank = os.path.join(cls.rdir, "pblank")
        cls.py_dis = os.path.join(cls.ppy_dis, "bin", "python")
        cls.py_blank = os.path.join(cls.ppy_blank, "bin", "python")
        # creating a distribute based install
        try:
            # `--no-site-packages` has been deprecated
            # https://virtualenv.pypa.io/en/stable/reference/#cmdoption-no-site-packages
            subprocess.check_call(
                [
                    salt.utils.path.which_bin(KNOWN_VIRTUALENV_BINARY_NAMES),
                    "--no-setuptools",
                    "--no-pip",
                    cls.ppy_dis,
                ]
            )
        except subprocess.CalledProcessError:
            subprocess.check_call(
                [salt.utils.path.which_bin(KNOWN_VIRTUALENV_BINARY_NAMES), cls.ppy_dis]
            )

            url = (
                "https://pypi.python.org/packages/source"
                "/d/distribute/distribute-0.6.43.tar.gz"
            )
            download_to(
                url,
                os.path.join(cls.ppy_dis, "distribute-0.6.43.tar.gz"),
            )

            subprocess.check_call(
                [
                    "tar",
                    "-C",
                    cls.ppy_dis,
                    "-xzvf",
                    f"{cls.ppy_dis}/distribute-0.6.43.tar.gz",
                ]
            )

            subprocess.check_call(
                [
                    f"{cls.ppy_dis}/bin/python",
                    f"{cls.ppy_dis}/distribute-0.6.43/setup.py",
                    "install",
                ]
            )

        # creating a blank based install
        try:
            subprocess.check_call(
                [
                    salt.utils.path.which_bin(KNOWN_VIRTUALENV_BINARY_NAMES),
                    "--no-setuptools",
                    "--no-pip",
                    cls.ppy_blank,
                ]
            )
        except subprocess.CalledProcessError:
            subprocess.check_call(
                [
                    salt.utils.path.which_bin(KNOWN_VIRTUALENV_BINARY_NAMES),
                    cls.ppy_blank,
                ]
            )

    @pytest.mark.skip(reason="TODO this test should probably be fixed")
    def test_buildout_bootstrap(self):
        b_dir = os.path.join(self.tdir, "b")
        bd_dir = os.path.join(self.tdir, "b", "bdistribute")
        b2_dir = os.path.join(self.tdir, "b", "b2")
        self.assertTrue(buildout._has_old_distribute(self.py_dis))
        # this is too hard to check as on debian & other where old
        # packages are present (virtualenv), we can't have
        # a clean site-packages
        # self.assertFalse(buildout._has_old_distribute(self.py_blank))
        self.assertFalse(buildout._has_old_distribute(self.py_st))
        self.assertFalse(buildout._has_setuptools7(self.py_dis))
        self.assertTrue(buildout._has_setuptools7(self.py_st))
        self.assertFalse(buildout._has_setuptools7(self.py_blank))

        ret = buildout.bootstrap(bd_dir, buildout_ver=1, python=self.py_dis)
        comment = ret["outlog"]
        self.assertTrue("--distribute" in comment)
        self.assertTrue("Generated script" in comment)

        ret = buildout.bootstrap(b_dir, buildout_ver=1, python=self.py_blank)
        comment = ret["outlog"]
        # as we may have old packages, this test the two
        # behaviors (failure with old setuptools/distribute)
        self.assertTrue(
            ("Got " in comment and "Generated script" in comment)
            or ("setuptools>=0.7" in comment)
        )

        ret = buildout.bootstrap(b_dir, buildout_ver=2, python=self.py_blank)
        comment = ret["outlog"]
        self.assertTrue(
            ("setuptools" in comment and "Generated script" in comment)
            or ("setuptools>=0.7" in comment)
        )

        ret = buildout.bootstrap(b_dir, buildout_ver=2, python=self.py_st)
        comment = ret["outlog"]
        self.assertTrue(
            ("setuptools" in comment and "Generated script" in comment)
            or ("setuptools>=0.7" in comment)
        )

        ret = buildout.bootstrap(b2_dir, buildout_ver=2, python=self.py_st)
        comment = ret["outlog"]
        self.assertTrue(
            ("setuptools" in comment and "Creating directory" in comment)
            or ("setuptools>=0.7" in comment)
        )

    @pytest.mark.slow_test
    def test_run_buildout(self):
        if salt.modules.virtualenv_mod.virtualenv_ver(self.ppy_st) >= (20, 0, 0):
            self.skipTest(
                "Skiping until upstream resolved"
                " https://github.com/pypa/virtualenv/issues/1715"
            )

        b_dir = os.path.join(self.tdir, "b")
        ret = buildout.bootstrap(b_dir, buildout_ver=2, python=self.py_st)
        self.assertTrue(ret["status"])
        ret = buildout.run_buildout(b_dir, parts=["a", "b"])
        out = ret["out"]
        self.assertTrue("Installing a" in out)
        self.assertTrue("Installing b" in out)

    @pytest.mark.slow_test
    def test_buildout(self):
        if salt.modules.virtualenv_mod.virtualenv_ver(self.ppy_st) >= (20, 0, 0):
            self.skipTest(
                "Skiping until upstream resolved"
                " https://github.com/pypa/virtualenv/issues/1715"
            )

        b_dir = os.path.join(self.tdir, "b")
        ret = buildout.buildout(b_dir, buildout_ver=2, python=self.py_st)
        self.assertTrue(ret["status"])
        out = ret["out"]
        comment = ret["comment"]
        self.assertTrue(ret["status"])
        self.assertIn("Creating directory", out)
        self.assertIn("Installing a.", out)
        self.assertTrue(f"{self.py_st} bootstrap.py" in comment)
        self.assertTrue("buildout -c buildout.cfg" in comment)
        ret = buildout.buildout(
            b_dir, parts=["a", "b", "c"], buildout_ver=2, python=self.py_st
        )
        outlog = ret["outlog"]
        out = ret["out"]
        comment = ret["comment"]
        self.assertIn("Installing single part: a", outlog)
        self.assertIn("buildout -c buildout.cfg -N install a", comment)
        self.assertIn("Installing b.", out)
        self.assertIn("Installing c.", out)
        ret = buildout.buildout(
            b_dir, parts=["a", "b", "c"], buildout_ver=2, newest=True, python=self.py_st
        )
        outlog = ret["outlog"]
        out = ret["out"]
        comment = ret["comment"]
        self.assertIn("buildout -c buildout.cfg -n install a", comment)


# TODO: Is this test even still needed?
class BuildoutAPITestCase(TestCase):
    def test_merge(self):
        buildout.LOG.clear()
        buildout.LOG.info("àé")
        buildout.LOG.info("àé")
        buildout.LOG.error("àé")
        buildout.LOG.error("àé")
        ret1 = buildout._set_status({}, out="éà")
        uret1 = buildout._set_status({}, out="éà")
        buildout.LOG.clear()
        buildout.LOG.info("ççàé")
        buildout.LOG.info("ççàé")
        buildout.LOG.error("ççàé")
        buildout.LOG.error("ççàé")
        ret2 = buildout._set_status({}, out="çéà")
        uret2 = buildout._set_status({}, out="çéà")
        uretm = buildout._merge_statuses([ret1, uret1, ret2, uret2])
        for ret in ret1, uret1, ret2, uret2:
            out = ret["out"]
            if not isinstance(ret["out"], str):
                out = ret["out"].decode("utf-8")

        for out in ["àé", "ççàé"]:
            self.assertIn(out, uretm["logs_by_level"]["info"])
            self.assertIn(out, uretm["outlog_by_level"])

    def test_setup(self):
        buildout.LOG.clear()
        buildout.LOG.info("àé")
        buildout.LOG.info("àé")
        buildout.LOG.error("àé")
        buildout.LOG.error("àé")
        ret = buildout._set_status({}, out="éà")
        uret = buildout._set_status({}, out="éà")
        self.assertTrue(ret["outlog"] == uret["outlog"])
        self.assertTrue("àé" in uret["outlog_by_level"])



"""
    Test Syslog returner

    :codeauthor: :email:`Megan Wilhite (mwilhite@saltstack.com)`
"""

import pytest

import salt.returners.syslog_return as syslog
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {syslog: {}}


@pytest.mark.skipif(not syslog.HAS_SYSLOG, reason="Skip when syslog not installed")
def test_syslog_returner_unicode():
    """
    test syslog returner with unicode
    """
    ret = {
        "fun_args": [],
        "jid": "20180713160901624786",
        "return": True,
        "retcode": 0,
        "success": True,
        "fun": "test.ping",
        "id": "02e10e971a30",
    }
    opts = {
        "level": "LOG_INFO",
        "options": [],
        "facility": "LOG_USER",
        "tag": "salt-minion",
    }

    with patch(
        "salt.returners.syslog_return._get_options", MagicMock(return_value=opts)
    ):
        try:
            syslog.returner(ret)
        except Exception as e:  # pylint: disable=broad-except
            pytest.fail(f"syslog.returner() failed with exception: {e}")



"""
Tests for atomicfile utility module.
"""

import pytest

import salt.utils.files
from salt.utils.atomicfile import atomic_open


@pytest.mark.skip_on_windows(reason="Not a Windows test")
def test_atomicfile_respects_umask(tmp_path):
    """
    Test that creating a file using atomic_open respects the umask, instead of
    creating the file with 0600 perms.
    """
    new_file = tmp_path / "foo"
    contents = "bar"

    # Set the umask specifically for this test so that we know what the mode of
    # the created file should be.
    with salt.utils.files.set_umask(0o022):
        with atomic_open(str(new_file), "w") as fh_:
            fh_.write(contents)

    assert new_file.read_text() == contents
    assert oct(new_file.stat().st_mode)[-3:] == "644"



"""
LGPO - Registry.pol
===================

.. versionadded:: 3006.0

A state module for working with registry based policies in Windows Local Group
Policy (LGPO). This module contains functions for working with the
``Registry.pol`` file. The ``Registry.pol`` file is the source of truth for
registry settings and LGPO.

Group Policy is refreshed every 90 seconds by default. During that refresh the
contents of the ``Registry.pol`` file are applied to the Registry. If the
setting is changed outside of Group Policy to something other than what is
contained in the ``Registry.pol`` file, it will be changed back during the next
refresh.

In the Group Policy Editor (``gpedit.msc``) these policies can be set to three
states:

- Not Configured
- Enabled
- Disabled

A policy that is "Not Configured" does not have an entry in the ``Registry.pol``
file. A Group Policy refresh will not make any changes to key/value pairs in the
registry that are not specified in the ``Registry.pol`` file.

An "Enabled" policy will have an entry in the ``Registry.pol`` files that
contains its key path, value name, value type, value size, and value data. When
Group Policy is refreshed, existing values will be overwritten with those
contained in the ``Registry.pol`` file.

A "Disabled" policy will have an entry in the ``Registry.pol`` file with the key
path and the value name, but the value name will be prepended with ``**del.``.
When Group Policy is refreshed the key/value will be deleted from the registry.
If the key contains no values, it will also be deleted.

Working with LGPO Reg
---------------------

The easiest way to figure out the values needed for this module is to set the
policy using the Group Policy Editor (``gpedit.msc``) and then run the
``lgpo_reg.read_reg_pol`` function. This function will display a dictionary of
all registry-based policies in the ``Registry.pol`` file. From its return you
can get the ``key``, ``v_name``, ``v_type``, and ``v_data`` required to
configure that policy.

.. note::
    Not all policies in the Group Policy Editor (``gpedit.msc``) that write to
    the registry make that change in the ``Registry.pol`` file. Those policies
    could still be enforced via the ``Registry.pol`` file... theoretically. But
    you will have to find the values needed to set them with this module using a
    different method.
"""

import salt.utils.data
import salt.utils.platform

__virtualname__ = "lgpo_reg"


def __virtual__():
    """
    Only works on Windows with the lgpo_reg module
    """
    if not salt.utils.platform.is_windows():
        return False, "LGPO_REG State: Only available on Windows"

    if "lgpo_reg.get_value" not in __salt__:
        return False, "LGPO_REG State: lgpo_reg module not available"

    return __virtualname__


def _get_current(key, name, policy_class):
    """
    Helper function to get the current state of the policy
    """
    hive = "HKLM"
    if policy_class == "User":
        hive = "HKCU"
    pol = __salt__["lgpo_reg.get_value"](
        key=key, v_name=name, policy_class=policy_class
    )
    reg_raw = __utils__["reg.read_value"](hive=hive, key=key, vname=name)

    reg = {}
    if reg_raw["vdata"] is not None:
        reg["data"] = reg_raw["vdata"]
    if reg_raw["vtype"] is not None:
        reg["type"] = reg_raw["vtype"]

    return {"pol": pol, "reg": reg}


def value_present(name, key, v_data, v_type="REG_DWORD", policy_class="Machine"):
    r"""
    Ensure a registry setting is present in the Registry.pol file.

    Args:

        name (str): The registry value name within the key

        key (str): The registry key path

        v_data(str): The registry value

        v_type (str): The registry value type. Must be one of the following:

            - REG_BINARY
            - REG_DWORD
            - REG_EXPAND_SZ
            - REG_MULTI_SZ
            - REG_QWORD
            - REG_SZ

            Default is REG_DWORD

        policy_class (str): The registry class to write to. Can be one of the
            following:

            - Computer
            - Machine
            - User

            Default is ``Machine``

    CLI Example:

    .. code-block:: yaml

        # Using the name parameter in the definition
        set_reg_pol_value:
          lgpo_reg.value_present:
            - key: SOFTWARE\MyKey
            - name: MyValue
            - v_type: REG_SZ
            - v_data: "some string data"
            - policy_class: Machine


        # Using the name as the parameter and modifying the User policy
        MyValue:
          lgpo_reg.value_present:
            - key: SOFTWARE\MyKey
            - v_type: REG_SZ
            - v_data: "some string data"
            - policy_class: User
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    old = _get_current(key=key, name=name, policy_class=policy_class)

    pol_correct = (
        str(old["pol"].get("data", "")) == str(v_data)
        and old["pol"].get("type", "") == v_type
    )
    reg_correct = (
        str(old["reg"].get("data", "")) == str(v_data)
        and old["reg"].get("type", "") == v_type
    )

    if pol_correct and reg_correct:
        ret["comment"] = "Policy value already present\nRegistry value already present"
        ret["result"] = True
        return ret

    if __opts__["test"]:
        if not pol_correct:
            ret["comment"] = "Policy value will be set"
        if not reg_correct:
            if ret["comment"]:
                ret["comment"] += "\n"
            ret["comment"] += "Registry value will be set"
        ret["result"] = None
        return ret

    __salt__["lgpo_reg.set_value"](
        key=key,
        v_name=name,
        v_data=v_data,
        v_type=v_type,
        policy_class=policy_class,
    )

    new = _get_current(key=key, name=name, policy_class=policy_class)

    pol_correct = (
        str(new["pol"]["data"]) == str(v_data) and new["pol"]["type"] == v_type
    )
    reg_correct = (
        str(new["reg"]["data"]) == str(v_data) and new["reg"]["type"] == v_type
    )

    if pol_correct and reg_correct:
        ret["comment"] = "Registry policy value has been set"
        ret["result"] = True
    elif not pol_correct:
        ret["comment"] = "Failed to set policy value"
    elif not reg_correct:
        if ret["comment"]:
            ret["comment"] += "\n"
        ret["comment"] += "Failed to set registry value"

    changes = salt.utils.data.recursive_diff(old, new)

    if changes:
        ret["changes"] = changes

    return ret


def value_disabled(name, key, policy_class="Machine"):
    r"""
    Ensure a registry setting is disabled in the Registry.pol file.

    Args:

        key (str): The registry key path

        name (str): The registry value name within the key

        policy_class (str): The registry class to write to. Can be one of the
            following:

            - Computer
            - Machine
            - User

            Default is ``Machine``

    CLI Example:

    .. code-block:: yaml

        # Using the name parameter in the definition
        set_reg_pol_value:
          lgpo_reg.value_disabled:
            - key: SOFTWARE\MyKey
            - name: MyValue
            - policy_class: Machine


        # Using the name as the parameter and modifying the User policy
        MyValue:
          lgpo_reg.value_disabled:
            - key: SOFTWARE\MyKey
            - policy_class: User
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    old = _get_current(key=key, name=name, policy_class=policy_class)

    pol_correct = old["pol"].get("data", "") == f"**del.{name}"
    reg_correct = old["reg"] == {}

    if pol_correct and reg_correct:
        ret["comment"] = "Registry policy value already disabled"
        ret["result"] = True
        return ret

    if __opts__["test"]:
        if not pol_correct:
            ret["comment"] = "Policy value will be disabled"
        if not reg_correct:
            if ret["comment"]:
                ret["comment"] += "\n"
            ret["comment"] += "Registry value will be removed"
        ret["result"] = None
        return ret

    __salt__["lgpo_reg.disable_value"](key=key, v_name=name, policy_class=policy_class)

    new = _get_current(key=key, name=name, policy_class=policy_class)

    pol_correct = new["pol"].get("data", "") == f"**del.{name}"
    reg_correct = new["reg"] == {}

    if pol_correct and reg_correct:
        ret["comment"] = "Registry policy value disabled"
        ret["result"] = True
    elif not pol_correct:
        ret["comment"] = "Failed to disable policy value"
    elif not reg_correct:
        if ret["comment"]:
            ret["comment"] += "\n"
        ret["comment"] += "Failed to remove registry value"

    changes = salt.utils.data.recursive_diff(old, new)

    if changes:
        ret["changes"] = changes

    return ret


def value_absent(name, key, policy_class="Machine"):
    r"""
    Ensure a registry setting is not present in the Registry.pol file.

    Args:

        key (str): The registry key path

        name (str): The registry value name within the key

        policy_class (str): The registry class to write to. Can be one of the
            following:

            - Computer
            - Machine
            - User

            Default is ``Machine``

    CLI Example:

    .. code-block:: yaml

        # Using the name parameter in the definition
        set_reg_pol_value:
          lgpo_reg.value_absent:
            - key: SOFTWARE\MyKey
            - name: MyValue
            - policy_class: Machine


        # Using the name as the parameter and modifying the User policy
        MyValue:
          lgpo_reg.value_absent:
            - key: SOFTWARE\MyKey
            - policy_class: User
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    old = _get_current(key=key, name=name, policy_class=policy_class)

    pol_correct = old["pol"] == {}
    reg_correct = old["reg"] == {}

    if pol_correct and reg_correct:
        ret["comment"] = "Registry policy value already deleted"
        ret["result"] = True
        return ret

    if __opts__["test"]:
        if not pol_correct:
            ret["comment"] = "Policy value will be deleted"
        if not reg_correct:
            if ret["comment"]:
                ret["comment"] += "\n"
            ret["comment"] += "Registry value will be deleted"
        ret["result"] = None
        return ret

    __salt__["lgpo_reg.delete_value"](key=key, v_name=name, policy_class=policy_class)

    new = _get_current(key=key, name=name, policy_class=policy_class)

    pol_correct = new["pol"] == {}
    reg_correct = new["reg"] == {}

    if pol_correct and reg_correct:
        ret["comment"] = "Registry policy value deleted"
        ret["result"] = True
    elif not pol_correct:
        ret["comment"] = "Failed to delete policy value"
    elif not reg_correct:
        if ret["comment"]:
            ret["comment"] += "\n"
        ret["comment"] += "Failed to delete registry value"

    changes = salt.utils.data.recursive_diff(old, new)

    if changes:
        ret["changes"] = changes

    return ret



import logging

import pytest

import salt.modules.deb_postgres as deb_postgres
from tests.support.mock import Mock, patch

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_unless_on_linux(reason="Only supported on Linux family"),
]


@pytest.fixture
def get_lscuster():
    return """\
8.4 main 5432 online postgres /srv/8.4/main \
        /var/log/postgresql/postgresql-8.4-main.log
9.1 main 5433 online postgres /srv/9.1/main \
        /var/log/postgresql/postgresql-9.1-main.log
"""


@pytest.fixture
def configure_loader_modules(get_lscuster):
    return {
        deb_postgres: {
            "__salt__": {
                "config.option": Mock(),
                "cmd.run_all": Mock(return_value={"stdout": get_lscuster}),
                "file.chown": Mock(),
                "file.remove": Mock(),
            }
        }
    }


def test_cluster_create():
    with patch("salt.utils.path.which", Mock(return_value="/usr/bin/pg_createcluster")):
        expected_cmdstr = (
            "/usr/bin/pg_createcluster "
            "--port 5432 --locale fr_FR --encoding UTF-8 "
            "--datadir /opt/postgresql "
            "9.3 main"
        )

        deb_postgres.cluster_create(
            "9.3",
            "main",
            port="5432",
            locale="fr_FR",
            encoding="UTF-8",
            datadir="/opt/postgresql",
        )
        assert deb_postgres.__salt__["cmd.run_all"].call_args[0][0] == expected_cmdstr


def test_cluster_create_with_initdb_options():
    with patch("salt.utils.path.which", Mock(return_value="/usr/bin/pg_createcluster")):
        expected_cmdstr = (
            "/usr/bin/pg_createcluster "
            "--port 5432 --locale fr_FR --encoding UTF-8 "
            "--datadir /opt/postgresql "
            "11 main "
            "-- "
            "--allow-group-access "
            "--data-checksums "
            "--wal-segsize 32"
        )

        deb_postgres.cluster_create(
            "11",
            "main",
            port="5432",
            locale="fr_FR",
            encoding="UTF-8",
            datadir="/opt/postgresql",
            allow_group_access=True,
            data_checksums=True,
            wal_segsize="32",
        )
        assert deb_postgres.__salt__["cmd.run_all"].call_args[0][0] == expected_cmdstr


def test_cluster_create_with_float():
    with patch("salt.utils.path.which", Mock(return_value="/usr/bin/pg_createcluster")):
        expected_cmdstr = (
            "/usr/bin/pg_createcluster "
            "--port 5432 --locale fr_FR --encoding UTF-8 "
            "--datadir /opt/postgresql "
            "9.3 main"
        )

        deb_postgres.cluster_create(
            9.3,
            "main",
            port="5432",
            locale="fr_FR",
            encoding="UTF-8",
            datadir="/opt/postgresql",
        )
        assert deb_postgres.__salt__["cmd.run_all"].call_args[0][0] == expected_cmdstr


def test_parse_pg_lsclusters(get_lscuster):
    with patch("salt.utils.path.which", Mock(return_value="/usr/bin/pg_lsclusters")):
        stdout = get_lscuster
        maxDiff = None
        expected = {
            "8.4/main": {
                "port": 5432,
                "status": "online",
                "user": "postgres",
                "datadir": "/srv/8.4/main",
                "log": "/var/log/postgresql/postgresql-8.4-main.log",
            },
            "9.1/main": {
                "port": 5433,
                "status": "online",
                "user": "postgres",
                "datadir": "/srv/9.1/main",
                "log": "/var/log/postgresql/postgresql-9.1-main.log",
            },
        }
        assert deb_postgres._parse_pg_lscluster(stdout) == expected


def test_cluster_list():
    with patch("salt.utils.path.which", Mock(return_value="/usr/bin/pg_lsclusters")):
        return_list = deb_postgres.cluster_list()
        assert (
            deb_postgres.__salt__["cmd.run_all"].call_args[0][0]
            == "/usr/bin/pg_lsclusters --no-header"
        )

        return_dict = deb_postgres.cluster_list(verbose=True)
        assert isinstance(return_dict, dict)


def test_cluster_exists():
    assert deb_postgres.cluster_exists("8.4")
    assert deb_postgres.cluster_exists("8.4", "main")
    assert not deb_postgres.cluster_exists("3.4", "main")


def test_cluster_delete():
    with patch("salt.utils.path.which", Mock(return_value="/usr/bin/pg_dropcluster")):
        deb_postgres.cluster_remove("9.3", "main")
        assert (
            deb_postgres.__salt__["cmd.run_all"].call_args[0][0]
            == "/usr/bin/pg_dropcluster 9.3 main"
        )

        deb_postgres.cluster_remove("9.3", "main", stop=True)
        assert (
            deb_postgres.__salt__["cmd.run_all"].call_args[0][0]
            == "/usr/bin/pg_dropcluster --stop 9.3 main"
        )

        deb_postgres.cluster_remove(9.3, "main", stop=True)
        assert (
            deb_postgres.__salt__["cmd.run_all"].call_args[0][0]
            == "/usr/bin/pg_dropcluster --stop 9.3 main"
        )
